""" Herre contrib package

This package contains contrib modules for the herre package. This allows for the
creation of herre powered links that can be used to dynamially retrieve a token
using an herre grant (e.g. an interactive grant) and send it to the next link.
"""
