""" Fakts contrib package

This package contains contrib modules for the fakts package.
It provides links that can be to connect to a graphql-server
whose endpoint is not known at "compile" time. (e.g. a server
whose ip-adress is dynamic). These links use the fakts protocol
to automatically discover the graphql endpoint.

"""
