""" Fakts contrib package

This package contains contrib modules for the fakts package. In this case, it
contains links that can be used to connect to a graphql-server whose endpoint
is not known at "compile" time. (e.g. a server whose ip-adress is dynamic).
These link uses the fakts protocol to automatically discover the graphql endpoint.
"""
